_("Show Desktop")
_("Exposé all windows")
_("Exposé current application")
_("Budgie Desktop Settings")
_("Lock screen")
_("Toggle Shuffler GUI")
_("All windows in grid")
_("Window to topleft")
_("Window to topright")
_("Window to bottomright")
_("Window to bottomleft")
_("Toggle Raven - AppletView")
_("Toggle Raven - Notifications")
