.trash-applet-header {
    border-bottom: 1px solid mix(@theme_base_color, #000000,  0.35);
    box-shadow: 0px 1px 1px alpha(@theme_fg_color, 0.04);
    font-weight: 500;
    padding-top: 10px;
    padding-bottom: 10px;
}

.trash-applet-list > row {
    border-bottom: 0.5px solid mix(@theme_base_color, #000000,  0.35);
}

.trash-applet-list > row:hover {
    background-color: transparent;
}

.trash-applet-list:last-child {
    border-bottom: none;
}

.trash-store-widget:hover {
    background-color: transparent;
}

.trash-store-header {
    padding: 0 5px;
}

.trash-store-header > revealer {
    border-bottom: 0.5px solid mix(@theme_base_color, #000000,  0.35);
    margin-top: 15px;
    margin-bottom: 15px;
}

.trash-file-box {
    background-color: transparent;
    border: 0.5px solid mix(@theme_base_color, #000000,  0.35);
    border-radius: 2px;
    margin-left: 5px;
    margin-right: 5px;
    margin-bottom: 15px;
}

.trash-file-box > row {
    transition: opacity, background-color 0.2s cubic-bezier(0.4, 0.0, 0.2, 1.0) 0.03s;
}

.trash-file-box.empty {
    border-bottom: none;
    padding-top: 0;
    padding-bottom: 0;
}

.trash-item {
    margin: 0;
}

.trash-info-revealer {
    padding-top: 5px;
    padding-bottom: 5px;
}

.trash-info-revealer > box > label {
    margin-left: 5px;
    margin-right: 5px;
}

.trash-settings-box {
    background-color: @theme_bg_color;
}

.trash-settings-box > box > label {
    border-bottom: 0.5px solid mix(@theme_base_color, #000000,  0.35);
    margin: 10px;
}

.trash-applet-footer {
    border-top: 1px solid mix(@theme_base_color, #000000,  0.35);
    padding-top: 5px;
    padding-bottom: 5px;
}
